/**
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2007</p>
 * <p>Company: </p>
 * @author Ioannis Partalas, Grigorios Tsoumakas 
 * @version 24/07/2007
 */


import weka.classifiers.*;
import weka.classifiers.bayes.*;
import weka.classifiers.functions.*;
import weka.classifiers.lazy.*;
import weka.classifiers.meta.*;
import weka.classifiers.rules.*;
import weka.classifiers.trees.*;
import weka.core.*;
import weka.experiment.*;
import weka.filters.*;
import weka.filters.unsupervised.attribute.*;
import weka.core.converters.*;
import java.io.*;
import java.util.*;
import weka.filters.supervised.instance.*;

public class Bagging
{
    public double TR;  // Percentage of the training set
    public double PR;  // Percentage of the pruning set
    public double TE;  // Percentage of the test set
    public int M;      // Number of models
    
    public J48 [] j48Model; // Models
    
    public Instances Train;
    public Instances Train80;
    public Instances Pruning;
    public Instances Test;

    public Bagging()
    {
    }

    private Instances LoadArff(String FileStem)
    {
	try {
	    ArffLoader aLoader = new ArffLoader();
	    FileInputStream aFile = new FileInputStream(FileStem);
	    aLoader.setSource(aFile);
	    Instances Data = new Instances(aLoader.getDataSet());
	    aFile.close();
	    return Data;
	}
	catch (Exception Ex) {
	    System.out.println(Ex);
	    System.out.println(FileStem);
	    return null;
	}
    }

    // Save an Instances abject in arff format
    private void SaveData(String FileStem,Instances data)
    {
        try
	{
	    File f = new File(FileStem);
	    FileWriter fw = new FileWriter(f);
	    fw.write(data.toString());
	    fw.close();
	}
	catch (Exception Ex){
	    System.out.println(Ex);
	    System.out.println(FileStem);
	}
    }


    public void CreateData(String dataset,String SourcePath, String TargetPath,Random rnd)
    {
	Instances Data = LoadArff(SourcePath+dataset+".arff");
	Data.setClassIndex(Data.numAttributes()-1);
	Data.randomize(rnd);


	int tr = (int)(TR*Data.numInstances());
	int pr = (int)(PR*Data.numInstances());
	int te = (int)(TE*Data.numInstances());

	System.out.println("TR:"+tr+" PR:" + pr +" TE:"+te);

	Train   = new Instances(Data, 1,tr);
	Pruning = new Instances(Data, tr, pr);
	Test    = new Instances(Data,tr+pr,te);
	Train80 = new Instances(Data,1,tr+pr);

	SaveData(TargetPath+dataset+"_train"+".arff",Train);
	SaveData(TargetPath+dataset+"_pruning"+".arff",Pruning);
	SaveData(TargetPath+dataset+"_test"+".arff",Test);
	SaveData(TargetPath+dataset+"_train80"+".arff",Train80);
	
	CreateModels(Train,TargetPath+"Models/");
        CreatePredictions(TargetPath+"Predictions/",false);
	CreateModels(Train80,TargetPath+"Models/");
        CreatePredictions(TargetPath+"Predictions/",true);
    }

    // Create M models through bagging on data and save them
    public void CreateModels(Instances data,String path)
    {
	Random rand  = new Random(System.currentTimeMillis());
	j48Model = new J48[M];

	try
	{
	    for(int j=0;j<M;j++)
	    {
		j48Model[j] = new J48();
		Resample rsampler = new Resample();
		rsampler.setRandomSeed(rand.nextInt());
		rsampler.setInputFormat(data);
		rsampler.setSampleSizePercent(100);

		Instances newInst = rsampler.useFilter(data,rsampler);
		j48Model[j].buildClassifier(newInst);

	//	ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(path+"tree"+j+".model"));
	//	oos.writeObject(j48Model[j]);
	//	oos.flush();
	//	oos.close();
	    }
	}
	catch(Exception Ex)
	{
	    System.out.println(Ex);
	}
    }

    public void CreatePredictions(String path,boolean type) 
    {
	for (int i=0; i<M; i++) 
	{
	    if(type)
	    {
		ResultsDataset predTrainData = new ResultsDataset();
		predTrainData.CreateDataset(j48Model[i],Train80,path+"train"+i+".pred");

		ResultsDataset predTestData = new ResultsDataset();
		predTestData.CreateDataset(j48Model[i],Test,path+"test"+i+".pred");
	    }
	    else
	    {
		ResultsDataset predPruneData = new ResultsDataset();
		predPruneData.CreateDataset(j48Model[i],Pruning,path+"prune"+i+".pred");

		ResultsDataset predTestData = new ResultsDataset();
		predTestData.CreateDataset(j48Model[i],Test,path+"testB"+i+".pred");
	    }
	}
    }
    
    public static void main(String[] args)
    {
	Bagging bagging = new Bagging();
	String source = null;
	String target = null;
	String file = null;
	Random rand = new Random(System.currentTimeMillis());

	bagging.TR = Double.parseDouble(args[0]);
	bagging.PR = Double.parseDouble(args[1]);
	bagging.TE = Double.parseDouble(args[2]);
	bagging.M = Integer.parseInt(args[3]);
	file = args[4];
	source = args[5];
	target = args[6];

	for(int i=1;i<=1;i++)
	    bagging.CreateData(file,source,target+"exp"+i+"/",rand);
    }
}
